@@ -3,28 +3,39 @@
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 
-
+from datetime import datetime
+import errno
 import io
 import os
 import shutil
+import stat
+import sys
+import warnings
 import mimetypes
 import nbformat
 
+from send2trash import send2trash
+from send2trash.exceptions import TrashPermissionError
 from tornado import web
 
 from .filecheckpoints import FileCheckpoints
 from .fileio import FileManagerMixin
 from .manager import ContentsManager
-
+from ...utils import exists
 
 from ipython_genutils.importstring import import_item
-from traitlets import Any, Unicode, Bool, TraitError
+from traitlets import Any, Unicode, Bool, TraitError, observe, default, validate
 from ipython_genutils.py3compat import getcwd, string_types
-from . import tz
+
+from notebook import _tz as tz
 from notebook.utils import (
-    is_hidden,
+    is_hidden, is_file_hidden,
     to_api_path,
 )
+from notebook.base.handlers import AuthenticatedFileHandler
+from notebook.transutils import _
+
+from os.path import samefile
 
 _script_exporter = None
 
@@ -32,9 +43,10 @@ _script_exporter = None
 def _post_save_script(model, os_path, contents_manager, **kwargs):
     """convert notebooks to Python script after save with nbconvert
 
-    replaces `ipython notebook --script`
+    replaces `jupyter notebook --script`
     """
     from nbconvert.exporters.script import ScriptExporter
+    warnings.warn("`_post_save_script` is deprecated and will be removed in Notebook 5.0", DeprecationWarning)
 
     if model['type'] != 'notebook':
         return
@@ -45,7 +57,6 @@ def _post_save_script(model, os_path, contents_manager, **kwargs):
     log = contents_manager.log
 
     base, ext = os.path.splitext(os_path)
-    py_fname = base + '.py'
     script, resources = _script_exporter.from_filename(os_path)
     script_fname = base + resources.get('output_extension', '.txt')
     log.info("Saving script /%s", to_api_path(script_fname, contents_manager.root_dir))
@@ -57,30 +68,36 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
 
     root_dir = Unicode(config=True)
 
-    def _root_dir_default(self):
+    @default('root_dir')
+    def _default_root_dir(self):
         try:
             return self.parent.notebook_dir
         except AttributeError:
             return getcwd()
 
-    save_script = Bool(False, config=True, help='DEPRECATED, use post_save_hook')
-    def _save_script_changed(self):
-        self.log.warn("""
-        `--script` is deprecated. You can trigger nbconvert via pre- or post-save hooks:
+    save_script = Bool(False, config=True, help='DEPRECATED, use post_save_hook. Will be removed in Notebook 5.0')
+    @observe('save_script')
+    def _update_save_script(self, change):
+        if not change['new']:
+            return
+        self.log.warning("""
+        `--script` is deprecated and will be removed in notebook 5.0.
+
+        You can trigger nbconvert via pre- or post-save hooks:
 
             ContentsManager.pre_save_hook
             FileContentsManager.post_save_hook
 
         A post-save hook has been registered that calls:
 
-            ipython nbconvert --to script [notebook]
+            jupyter nbconvert --to script [notebook]
 
         which behaves similarly to `--script`.
         """)
 
         self.post_save_hook = _post_save_script
 
-    post_save_hook = Any(None, config=True,
+    post_save_hook = Any(None, config=True, allow_none=True,
         help="""Python callable or importstring thereof
 
         to be called on the path of a file just saved.
@@ -97,12 +114,15 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
         - contents_manager: this ContentsManager instance
         """
     )
-    def _post_save_hook_changed(self, name, old, new):
-        if new and isinstance(new, string_types):
-            self.post_save_hook = import_item(self.post_save_hook)
-        elif new:
-            if not callable(new):
-                raise TraitError("post_save_hook must be callable")
+
+    @validate('post_save_hook')
+    def _validate_post_save_hook(self, proposal):
+        value = proposal['value']
+        if isinstance(value, string_types):
+            value = import_item(value)
+        if not callable(value):
+            raise TraitError("post_save_hook must be callable")
+        return value
 
     def run_post_save_hook(self, model, os_path):
         """Run the post-save hook if defined, and log errors"""
@@ -110,21 +130,39 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
             try:
                 self.log.debug("Running post-save hook on %s", os_path)
                 self.post_save_hook(os_path=os_path, model=model, contents_manager=self)
-            except Exception:
-                self.log.error("Post-save hook failed on %s", os_path, exc_info=True)
+            except Exception as e:
+                self.log.error("Post-save hook failed o-n %s", os_path, exc_info=True)
+                raise web.HTTPError(500, u'Unexpected error while running post hook save: %s'
+                                    % e) from e
 
-    def _root_dir_changed(self, name, old, new):
+    @validate('root_dir')
+    def _validate_root_dir(self, proposal):
         """Do a bit of validation of the root_dir."""
-        if not os.path.isabs(new):
+        value = proposal['value']
+        if not os.path.isabs(value):
             # If we receive a non-absolute path, make it absolute.
-            self.root_dir = os.path.abspath(new)
-            return
-        if not os.path.isdir(new):
-            raise TraitError("%r is not a directory" % new)
+            value = os.path.abspath(value)
+        if not os.path.isdir(value):
+            raise TraitError("%r is not a directory" % value)
+        return value
 
+    @default('checkpoints_class')
     def _checkpoints_class_default(self):
         return FileCheckpoints
 
+    delete_to_trash = Bool(True, config=True,
+        help="""If True (default), deleting files will send them to the
+        platform's trash/recycle bin, where they can be recovered. If False,
+        deleting files really deletes them.""")
+
+    @default('files_handler_class')
+    def _files_handler_class_default(self):
+        return AuthenticatedFileHandler
+
+    @default('files_handler_params')
+    def _files_handler_params_default(self):
+        return {'path': self.root_dir}
+
     def is_hidden(self, path):
         """Does the API style path correspond to a hidden directory or file?
 
@@ -199,14 +237,36 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
         """
         path = path.strip('/')
         os_path = self._get_os_path(path=path)
-        return os.path.exists(os_path)
+        return exists(os_path)
 
     def _base_model(self, path):
         """Build the common base of a contents model"""
         os_path = self._get_os_path(path)
-        info = os.stat(os_path)
-        last_modified = tz.utcfromtimestamp(info.st_mtime)
-        created = tz.utcfromtimestamp(info.st_ctime)
+        info = os.lstat(os_path)
+        
+        try:
+            # size of file 
+            size = info.st_size
+        except (ValueError, OSError):
+            self.log.warning('Unable to get size.')
+            size = None
+        
+        try:
+            last_modified = tz.utcfromtimestamp(info.st_mtime)
+        except (ValueError, OSError):
+            # Files can rarely have an invalid timestamp
+            # https://github.com/jupyter/notebook/issues/2539
+            # https://github.com/jupyter/notebook/issues/2757
+            # Use the Unix epoch as a fallback so we don't crash.
+            self.log.warning('Invalid mtime %s for %s', info.st_mtime, os_path)
+            last_modified = datetime(1970, 1, 1, 0, 0, tzinfo=tz.UTC)
+
+        try:
+            created = tz.utcfromtimestamp(info.st_ctime)
+        except (ValueError, OSError):  # See above
+            self.log.warning('Invalid ctime %s for %s', info.st_ctime, os_path)
+            created = datetime(1970, 1, 1, 0, 0, tzinfo=tz.UTC)
+
         # Create the base model.
         model = {}
         model['name'] = path.rsplit('/', 1)[-1]
@@ -216,6 +276,8 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
         model['content'] = None
         model['format'] = None
         model['mimetype'] = None
+        model['size'] = size
+
         try:
             model['writable'] = os.access(os_path, os.W_OK)
         except OSError:
@@ -234,7 +296,7 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
 
         if not os.path.isdir(os_path):
             raise web.HTTPError(404, four_o_four)
-        elif is_hidden(os_path, self.root_dir):
+        elif is_hidden(os_path, self.root_dir) and not self.allow_hidden:
             self.log.info("Refusing to serve hidden directory %r, via 404 Error",
                 os_path
             )
@@ -242,28 +304,53 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
 
         model = self._base_model(path)
         model['type'] = 'directory'
+        model['size'] = None
         if content:
             model['content'] = contents = []
             os_dir = self._get_os_path(path)
             for name in os.listdir(os_dir):
-                os_path = os.path.join(os_dir, name)
-                # skip over broken symlinks in listing
-                if not os.path.exists(os_path):
-                    self.log.warn("%s doesn't exist", os_path)
+                try:
+                    os_path = os.path.join(os_dir, name)
+                except UnicodeDecodeError as e:
+                    self.log.warning(
+                        "failed to decode filename '%s': %s", name, e)
                     continue
-                elif not os.path.isfile(os_path) and not os.path.isdir(os_path):
+
+                try:
+                    st = os.lstat(os_path)
+                except OSError as e:
+                    # skip over broken symlinks in listing
+                    if e.errno == errno.ENOENT:
+                        self.log.warning("%s doesn't exist", os_path)
+                    else:
+                        self.log.warning("Error stat-ing %s: %s", os_path, e)
+                    continue
+
+                if (not stat.S_ISLNK(st.st_mode)
+                        and not stat.S_ISREG(st.st_mode)
+                        and not stat.S_ISDIR(st.st_mode)):
                     self.log.debug("%s not a regular file", os_path)
                     continue
-                if self.should_list(name) and not is_hidden(os_path, self.root_dir):
-                    contents.append(self.get(
-                        path='%s/%s' % (path, name),
-                        content=False)
-                    )
 
+                try:
+                    if self.should_list(name):
+                        if self.allow_hidden or not is_file_hidden(os_path, stat_res=st):
+                            contents.append(
+                                    self.get(path='%s/%s' % (path, name), content=False)
+                            )
+                except OSError as e:
+                    # ELOOP: recursive symlink
+                    if e.errno != errno.ELOOP:
+                        self.log.warning(
+                            "Unknown error checking if file %r is hidden",
+                            os_path,
+                            exc_info=True,
+                        )
             model['format'] = 'json'
 
         return model
 
+
     def _file_model(self, path, content=True, format=None):
         """Build a model for a file
 
@@ -304,13 +391,15 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
         """
         model = self._base_model(path)
         model['type'] = 'notebook'
+        os_path = self._get_os_path(path)
+        
         if content:
-            os_path = self._get_os_path(path)
             nb = self._read_notebook(os_path, as_version=4)
             self.mark_trusted_cells(nb, path)
             model['content'] = nb
             model['format'] = 'json'
             self.validate_notebook_model(model)
+            
         return model
 
     def get(self, path, content=True, type=None, format=None):
@@ -357,7 +446,7 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
 
     def _save_directory(self, os_path, model, path=''):
         """create a directory"""
-        if is_hidden(os_path, self.root_dir):
+        if is_hidden(os_path, self.root_dir) and not self.allow_hidden:
             raise web.HTTPError(400, u'Cannot create hidden directory %r' % os_path)
         if not os.path.exists(os_path):
             with self.perm_to_403():
@@ -400,7 +489,8 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
             raise
         except Exception as e:
             self.log.error(u'Error while saving file: %s %s', path, e, exc_info=True)
-            raise web.HTTPError(500, u'Unexpected error while saving file: %s %s' % (path, e))
+            raise web.HTTPError(500, u'Unexpected error while saving file: %s %s' %
+                                (path, e)) from e
 
         validation_message = None
         if model['type'] == 'notebook':
@@ -420,19 +510,35 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
         path = path.strip('/')
         os_path = self._get_os_path(path)
         rm = os.unlink
-        if os.path.isdir(os_path):
-            listing = os.listdir(os_path)
-            # Don't delete non-empty directories.
-            # A directory containing only leftover checkpoints is
-            # considered empty.
-            cp_dir = getattr(self.checkpoints, 'checkpoint_dir', None)
-            for entry in listing:
-                if entry != cp_dir:
-                    raise web.HTTPError(400, u'Directory %s not empty' % os_path)
-        elif not os.path.isfile(os_path):
-            raise web.HTTPError(404, u'File does not exist: %s' % os_path)
+        if not os.path.exists(os_path):
+            raise web.HTTPError(404, u'File or directory does not exist: %s' % os_path)
+
+        def is_non_empty_dir(os_path):
+            if os.path.isdir(os_path):
+                # A directory containing only leftover checkpoints is
+                # considered empty.
+                cp_dir = getattr(self.checkpoints, 'checkpoint_dir', None)
+                if set(os.listdir(os_path)) - {cp_dir}:
+                    return True
+
+            return False
+
+        if self.delete_to_trash:
+            if sys.platform == 'win32' and is_non_empty_dir(os_path):
+                # send2trash can really delete files on Windows, so disallow
+                # deleting non-empty files. See Github issue 3631.
+                raise web.HTTPError(400, u'Directory %s not empty' % os_path)
+            try:
+                self.log.debug("Sending %s to trash", os_path)
+                send2trash(os_path)
+                return
+            except TrashPermissionError as e:
+                self.log.warning("Skipping trash for %s, %s", os_path, e)
 
         if os.path.isdir(os_path):
+            # Don't permanently delete non-empty directories.
+            if is_non_empty_dir(os_path):
+                raise web.HTTPError(400, u'Directory %s not empty' % os_path)
             self.log.debug("Removing directory %s", os_path)
             with self.perm_to_403():
                 shutil.rmtree(os_path)
@@ -448,11 +554,15 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
         if new_path == old_path:
             return
 
+        # Perform path validation prior to converting to os-specific value since this
+        # is still relative to root_dir.
+        self._validate_path(new_path)
+
         new_os_path = self._get_os_path(new_path)
         old_os_path = self._get_os_path(old_path)
 
         # Should we proceed with the move?
-        if os.path.exists(new_os_path):
+        if os.path.exists(new_os_path) and not samefile(old_os_path, new_os_path):
             raise web.HTTPError(409, u'File already exists: %s' % new_path)
 
         # Move the file
@@ -462,15 +572,38 @@ class FileContentsManager(FileManagerMixin, ContentsManager):
         except web.HTTPError:
             raise
         except Exception as e:
-            raise web.HTTPError(500, u'Unknown error renaming file: %s %s' % (old_path, e))
+            raise web.HTTPError(500, u'Unknown error renaming file: %s %s' %
+                                (old_path, e)) from e
 
     def info_string(self):
-        return "Serving notebooks from local directory: %s" % self.root_dir
+        return _("Serving notebooks from local directory: %s") % self.root_dir
 
     def get_kernel_path(self, path, model=None):
         """Return the initial API path of  a kernel associated with a given notebook"""
+        if self.dir_exists(path):
+            return path
         if '/' in path:
             parent_dir = path.rsplit('/', 1)[0]
         else:
             parent_dir = ''
         return parent_dir
+
+    @staticmethod
+    def _validate_path(path):
+        """Checks if the path contains invalid characters relative to the current platform"""
+
+        if sys.platform == 'win32':
+            # On Windows systems, we MUST disallow colons otherwise an Alternative Data Stream will
+            # be created and confusion will reign! (See https://github.com/jupyter/notebook/issues/5190)
+            # Go ahead and add other invalid (and non-path-separator) characters here as well so there's
+            # consistent behavior - although all others will result in '[Errno 22]Invalid Argument' errors.
+            invalid_chars = '?:><*"|'
+        else:
+            # On non-windows systems, allow the underlying file creation to perform enforcement when appropriate
+            invalid_chars = ''
+
+        for char in invalid_chars:
+            if char in path:
+                raise web.HTTPError(400, "Path '{}' contains characters that are invalid for the filesystem. "
+                                         "Path names on this filesystem cannot contain any of the following "
+                                         "characters: {}".format(path, invalid_chars))
