@@ -1,46 +1,76 @@
 // Copyright (c) Jupyter Development Team.
 // Distributed under the terms of the Modified BSD License.
 
-require([
+// adapted from Mozilla Developer Network example at
+// https://developer.mozilla.org/en/JavaScript/Reference/Global_Objects/Function/bind
+// shim `bind` for testing under casper.js
+var bind = function bind(obj) {
+  var slice = [].slice;
+  var args = slice.call(arguments, 1),
+    self = this,
+    nop = function() {
+    },
+    bound = function() {
+      return self.apply(this instanceof nop ? this : (obj || {}), args.concat(slice.call(arguments)));
+    };
+  nop.prototype = this.prototype || {}; // Firefox cries sometimes if prototype is undefined
+  bound.prototype = new nop();
+  return bound;
+};
+Function.prototype.bind = Function.prototype.bind || bind ;
+
+
+requirejs([
     'jquery',
+    'contents',
     'base/js/namespace',
     'base/js/dialog',
     'base/js/events',
+    'base/js/promises',
     'base/js/page',
     'base/js/utils',
     'services/config',
-    'contents',
     'tree/js/notebooklist',
     'tree/js/sessionlist',
     'tree/js/kernellist',
     'tree/js/terminallist',
     'tree/js/newnotebook',
+    'tree/js/shutdownbutton',
     'auth/js/loginwidget',
-    // only loaded, not used:
-    'jqueryui',
-    'bootstrap',
-    'custom/custom',
+    'bidi/bidi',
 ], function(
     $,
+    contents_service,
     IPython,
     dialog,
     events,
+    promises,
     page,
     utils,
     config,
-    contents_service,
     notebooklist,
     sesssionlist,
     kernellist,
     terminallist,
     newnotebook,
-    loginwidget){
+    shutdownbutton,
+    loginwidget,
+    bidi){
     "use strict";
-
-    IPython.NotebookList = notebooklist.NotebookList;
-
-    page = new page.Page();
     
+    try{
+        requirejs(['custom/custom'], function() {});
+        bidi.loadLocale();
+    } catch(err) {
+        console.log("Error loading custom.js from tree service. Continuing and logging");
+        console.warn(err);
+    }
+
+    console.log('Welcome to Project Jupyter! Explore the various tools available and their corresponding documentation. If you are interested in contributing to the platform, please visit the community resources section at http://jupyter.org/community.html.');
+
+
+    // Setup all of the config related things
+
     var common_options = {
         base_url: utils.get_body_data("baseUrl"),
         notebook_path: utils.get_body_data("notebookPath"),
@@ -51,6 +81,10 @@ require([
     var common_config = new config.ConfigSection('common', common_options);
     common_config.load();
 
+    // Instantiate the main objects
+
+    page = new page.Page('div#header', 'div#site');
+
     var session_list = new sesssionlist.SesssionList($.extend({
         events: events},
         common_options));
@@ -58,12 +92,13 @@ require([
         base_url: common_options.base_url,
         common_config: common_config
     });
+    IPython.NotebookList = notebooklist.NotebookList;
     var notebook_list = new notebooklist.NotebookList('#notebook_list', $.extend({
         contents: contents,
-        session_list:  session_list}, 
+        session_list:  session_list},
         common_options));
     var kernel_list = new kernellist.KernelList('#running_list',  $.extend({
-        session_list:  session_list}, 
+        session_list:  session_list},
         common_options));
     
     var terminal_list;
@@ -75,7 +110,7 @@ require([
 
     var new_buttons = new newnotebook.NewNotebookWidget("#new-buttons",
         $.extend(
-            {contents: contents},
+            {contents: contents, events: events},
             common_options
         )
     );
@@ -137,7 +172,7 @@ require([
 
     page.show();
 
-    // For backwards compatability.
+    // For backwards compatibility.
     IPython.page = page;
     IPython.notebook_list = notebook_list;
     IPython.session_list = session_list;
@@ -146,6 +181,8 @@ require([
     IPython.new_notebook_widget = new_buttons;
 
     events.trigger('app_initialized.DashboardApp');
+    
+    // Now actually load nbextensions
     utils.load_extensions_from_config(cfg);
     utils.load_extensions_from_config(common_config);
     
@@ -153,6 +190,15 @@ require([
     $("#alternate_upload").change(function (event){
         notebook_list.handleFilesUpload(event,'form');
     });
+
+    // bound the the span around the input file upload to enable keyboard click
+    $("#upload_span").keydown(function (event) {
+        var key = event.which;
+        if ((key === 13) || (key === 32)) {
+            event.preventDefault();
+            $("#upload_span_input").click();
+        }
+    })
     
     // set hash on tab click
     $("#tabs").find("a").click(function(e) {
@@ -161,7 +207,7 @@ require([
         e.preventDefault();
 
         // Set the hash without causing the page to jump.
-        // http://stackoverflow.com/a/14690177/2824256
+        // https://stackoverflow.com/a/14690177/2824256
         var hash = $(this).attr("href");
         if(window.history.pushState) {
             window.history.pushState(null, null, hash);
@@ -172,6 +218,8 @@ require([
     
     // load tab if url hash
     if (window.location.hash) {
-        $("#tabs").find("a[href=" + window.location.hash + "]").click();
+        $("#tabs").find("a[href='" + window.location.hash + "']").click();
     }
+    
+    shutdownbutton.activate();
 });
